/**
 * Created by janis dähne.
 */

import * as React from "react";
import {connect} from "react-redux";
import {bindActionCreators, Dispatch} from "redux";
import {returntypeof} from 'react-redux-typescript';
import {RootState} from "../../../../../state/reducers";
import {Form, Icon, Tab} from "semantic-ui-react";
import MaterialInput from '../../../../material/materialInput'
import MaterialSelectWrapper from '../../../../material/materialSelectWrapper'
import {validationRules} from "../../../../../state/reducers/exerciseEditorSite/dialogs/changeTestDialog/editorTestReducer";
import {validationMessageKeys} from "../../../../../state/reducers/exerciseEditorSite/dialogs/changeTestDialog/editorTestReducerValidation";
import SimpleCodeEditorWrapper from '../../../../codeEditors/simpleCodeEditorWrapper'

import ConsolePanelView from '../../../doExerciseSite/consolePanel/consolePanelView'
import {TestProtocolItem} from "../../../../../types/testResults";
import TestFilesListView from '../../../exerciseEditorSite/exerciseTestsPanel/dialogs/testFilesListView'
import {CompareFileTestContentFrontendOnly} from "../../../../../types/exerciseEditor";
import {EditorTab} from "../../../../codeEditors/multiTabCodeEditorProps";
import {SelectOnChangeEventData} from "../../../../../types/reactEvents";
import {convertTestProtocolFromString, parseId} from "../../../../../helpers/convertersAndTransformers";
import Logger from "../../../../../helpers/logger";
import {KnownInternalTestTypes} from "../../../../../types/testTypes";
import {
  checkRegexs,
  convertCompareTestToObj,
  convertRegexTestObject,
  parseRegexTestContent,
  RegexTestObject,
  RegexTestObjectConverted
} from "../../../../../types/testsAndTestProtocol";
import {isNotEmptyOrSpaces} from "../../../../../validation/validations";
import {
  aceCompareFilesTestLanguage,
  aceRegexLanguage,
  aceTestProtocolLanguage,
  aceTextLanguage
} from "../../../../../helpers/editorHelper";
import {errorDialog} from "../../../../../helpers/dialogHelper";
import {
  setDisplayNameAction,
  setEditorTest_compileTimeoutInMs,
  setEditorTest_files,
  setEditorTest_maxDiskSpaceInKb,
  setEditorTest_memoryLimitInKb,
  setEditorTest_timeoutInMs,
  setEditorTest_compilerOptions,
  setEditorTestContent,
  setTestTypeId,
  setWeight
} from "../../../../../state/actions/editCustomProjectSite/dialogs/changeCustomProjectTestDialog/customProjectTestActions";
import {setSelectedTabIndex} from "../../../../../state/actions/editCustomProjectSite/dialogs/changeCustomProjectTestDialog/changeCustomProjectTestDialogActions";
import {RequestApiPrefix} from "../../../../../communicationLayer/constants";
import {getI18n} from "../../../../../../i18n/i18nRoot";
import {descriptionEditorDebounceInMs, TestTypeLimits} from '../../../../../constants'
import CompareFilesTestSettingsView
  from '../../../exerciseEditorSite/exerciseTestsPanel/dialogs/compareFilesTestSettingsView'
import TestSettingsView from '../../../../helpers/tests/testSettingsView'
import ManageAssetFilesModal from '../../../exerciseEditorSite/addAssetFilesModal/addAssetFilesModal'
import {
  setAddCustomProjectCustomTestAssetFilesModalFilterText,
  setAddCustomProjectCustomTestAssetFilesModalUploadPercentage,
  setIsAddCustomProjectCustomTestAssetFilesModalDisplayed
} from '../../../../../state/actions/editCustomProjectSite/editCustomProjectActions'
import axios, {CancelTokenSource} from 'axios'
import {
  attachExitingFileToCustomProjectCustomTestAsync,
  detachFileFromCustomProjectCustomTestAsync
} from '../../../../../state/actions/editCustomProjectSite/editCustomProjectCrudActions'
import {getPossibleAllUserFilesAsync} from '../../../../../state/actions/doExerciseSite/doExerciseCrudActions'
import {uploadCustomProjectCustomTestAssetFile} from '../../../../../communicationLayer/editCustomProjectLayer'
import {setCustomProjectTests} from '../../../../../state/actions/editCustomProjectSite/subSet/customProjectTestActions'
import {HelpPopup} from '../../../../helpers/helpPopup'

//import jsonlint from 'jsonlint'

//const css = require('./styles.styl');

export interface MyProps {
  //readonly test: string
  readonly className?: string
}

const mapStateToProps = (rootState: RootState, props: MyProps) => {
  return {
    //test0: rootState...
    //test: props.test
    ...props,
    test: rootState.changeCustomProjectTestDialogState.test,
    testTypes: rootState.testTypesState.testTypes,
    selectedTabIndex: rootState.changeCustomProjectTestDialogState.selectedTabIndex,
    customProjectId: rootState.editCustomProjectSiteState.customProject.id,

    allCustomProjectCustomTests: rootState.editCustomProjectSiteState.customProject.tests,

    isAddCustomProjectCustomTestAssetFilesModalDisplayed: rootState.editCustomProjectSiteState.isAddCustomProjectCustomTestAssetFilesModalDisplayed,
    isAddCustomProjectCustomTestAssetFilesModalBusy: rootState.editCustomProjectSiteState.isAddCustomProjectCustomTestAssetFilesModalBusy,
    addCustomProjectCustomTestAssetFilesModalFilterText: rootState.editCustomProjectSiteState.addCustomProjectCustomTestAssetFilesModalFilterText,
    addCustomProjectCustomTestAssetFilesModalUploadPercentage: rootState.editCustomProjectSiteState.addCustomProjectCustomTestAssetFilesModalUploadPercentage,
    allCustomProjectAssetFilePreviews: rootState.editCustomProjectSiteState.allCustomProjectAssetFilePreviews,

    langId: rootState.i18nState.langId
  }
}

const mapDispatchToProps = (dispatch: Dispatch) => bindActionCreators({
                                                                        //imported reducer funcs here
                                                                        setDisplayNameAction,
                                                                        setWeight,
                                                                        setTestTypeId,
                                                                        setEditorTestContent,

                                                                        setEditorTest_memoryLimitInKb,
                                                                        setEditorTest_timeoutInMs,
                                                                        setEditorTest_compileTimeoutInMs,
                                                                        setEditorTest_maxDiskSpaceInKb,
                                                                        setEditorTest_compilerOptions,

                                                                        setSelectedTabIndex,
                                                                        setEditorTest_files,

                                                                        setIsAddCustomProjectCustomTestAssetFilesModalDisplayed,
                                                                        setAddCustomProjectCustomTestAssetFilesModalFilterText,
                                                                        setAddCustomProjectCustomTestAssetFilesModalUploadPercentage,

                                                                        getPossibleAllUserFilesAsync,
                                                                        attachExitingFileToCustomProjectCustomTestAsync,
                                                                        detachFileFromCustomProjectCustomTestAsync,

                                                                        setCustomProjectTests,

                                                                      }, dispatch)


const stateProps = returntypeof(mapStateToProps);
const dispatchProps = returntypeof(mapDispatchToProps);
type Props = typeof stateProps & typeof dispatchProps;


const testAssetInputId = 'testAssetInputId'

export const changeTestViewCustomProjectEditorId = 'changeCustomProjectTestView'

const ChangeTestViewResizeId = 'ChangeTestViewResizeId'

// tslint:disable-next-line
let currentFileUploadCancelToken: CancelTokenSource | null = null

class ChangeTestView extends React.Component<Props, any> {

  componentWillMount(): void {

    if (this.props.test.testTypeId <= 0 && this.props.testTypes.length > 0) {
      //not selected ... select the first one
      this.props.setTestTypeId(this.props.testTypes[0].id)
    }
  }

  render(): JSX.Element {

    const testType = this.props.testTypes.find(p => p.id === this.props.test.testTypeId)

    if (this.props.test.testTypeId <= 0 || !testType) {
      //this happens when we render and we display the initial value (test where testTypeId is negative)
      //in componentWillMount we set the test type id and we should be finde on the second render
      return null
    }

    const isCompareFilesTest = testType.internalName === KnownInternalTestTypes.compareFilesTest

    const isRegexTest = testType.internalName === KnownInternalTestTypes.regexTest

    // noinspection TsLint
    let testParseError = '' //error when parsing the json object
    // noinspection TsLint
    let regexError = '' //error when a regex was invalid

    if (isRegexTest) {

      // noinspection TsLint
      let result: RegexTestObject | string = null
      // noinspection TsLint
      let convertedResult: RegexTestObjectConverted = null

      try {
        result = parseRegexTestContent(this.props.test.content)

        if (typeof result === 'string') {
          throw new Error(result)
        }

        convertedResult = convertRegexTestObject(result)
      } catch (err) {
        testParseError = err.message
      }

      if (convertedResult === null) {

        if (isNotEmptyOrSpaces(testParseError) === false) { //if testParseError is empty
          regexError = 'some regex are not properly separated, check that every regex is enclosed in /.../\n' +
            `also check that you don't use && or || inside a regex`
        }

      } else {

        const errors = checkRegexs(convertedResult)

        if (errors && errors.length > 0) {
          regexError = errors.join('\n')
        } else {


        }
      }
    }

    //we fake a test result to reuse the console view
    const visualTest: ReadonlyArray<TestProtocolItem> | null = convertTestProtocolFromString(this.props.test.content,
                                                                                             testType
    )


    const tempEditorTab: EditorTab = {
      id: this.props.test.id,
      content: this.props.test.content,
      language: testType.internalName === KnownInternalTestTypes.regexTest
                ? aceRegexLanguage
                : testType.internalName === KnownInternalTestTypes.blackBoxTest
                  ? aceTestProtocolLanguage
                  : testType.internalName === KnownInternalTestTypes.compareFilesTest
                    ? aceCompareFilesTestLanguage
                    : aceTextLanguage,
      isReadonly: false,
      fileNameWithExtensions: this.props.test.displayName,
      displayIndex: this.props.test.displayIndex,
      isMainFile: true,
      isContentVisibleForUser: true,
    }

    // tslint:disable-next-line
    let fileCompares: CompareFileTestContentFrontendOnly[] = []

    try {
      if (isCompareFilesTest) {
        const temp = this.props.test.content
          .split('\n')
          .filter(p => p !== '')
          .map<CompareFileTestContentFrontendOnly | null>(convertCompareTestToObj)
          .filter(p => p !== null)
        fileCompares = temp
      }
    } catch (err) {
      //TODO show error??
    }

    return (
      <div className="fh flexed-h">
        <div className="ui form divided-checkboxes fh" style={{height: 'auto' /*firefox fix*/}}>


          <ManageAssetFilesModal
            title={getI18n(this.props.langId, "Manage asset files")}
            isDisplayed={this.props.isAddCustomProjectCustomTestAssetFilesModalDisplayed}
            isBusy={this.props.isAddCustomProjectCustomTestAssetFilesModalBusy}
            uploadPercentage={this.props.addCustomProjectCustomTestAssetFilesModalUploadPercentage}
            uploadInputId={'custom-project-custom-test-upload'}
            onUploadFile={async (file) => {

              this.props.setIsAddCustomProjectCustomTestAssetFilesModalDisplayed(true, true)
              this.props.setAddCustomProjectCustomTestAssetFilesModalUploadPercentage(0)

              currentFileUploadCancelToken = axios.CancelToken.source()

              try {
                const assetFilePreview = await uploadCustomProjectCustomTestAssetFile(this.props.customProjectId,
                                                                                      this.props.test.id,
                                                                                      file,
                                                                                      currentFileUploadCancelToken.token,
                                                                                      this.props.setAddCustomProjectCustomTestAssetFilesModalUploadPercentage
                )

                currentFileUploadCancelToken = null

                //hide progress bar
                this.props.setAddCustomProjectCustomTestAssetFilesModalUploadPercentage(null)
                Logger.debug('editCustomProjectSite...changeTestView',
                             `upload finished (new id: ${assetFilePreview.id}), now downloading file content again...`
                )

                //this is a bit dirty... because it's not directly connected to the curd action... but ok for now (we call the raw crud operation and not a redux action)

                //set for the whole custom project!
                this.props.setCustomProjectTests(
                  this.props.allCustomProjectCustomTests.map(p => p.id !== this.props.test.id
                                                                  ? p
                                                                  : {
                      ...p,
                      files: p.files.concat(assetFilePreview)
                    }))

                //because this dialog edits a copy of the exercise we need to add the new preview there too
                this.props.setEditorTest_files(this.props.test.files.concat(assetFilePreview))


                //renew all asset preview list (could have changed) this is easier than to manage all special cases
                //e.g. detach file... was it deleted or just detached??
                // this.props.finishAddAssetFilesModalUpload(assetFilePreview)
                await this.props.getPossibleAllUserFilesAsync()

              } catch (err) {
                Logger.error('could not upload file, err: ' + err.message)
              } finally {
                currentFileUploadCancelToken = null
                this.props.setIsAddCustomProjectCustomTestAssetFilesModalDisplayed(true, false)
                this.props.setAddCustomProjectCustomTestAssetFilesModalUploadPercentage(null)
              }

            }}

            onUploadCanceled={
              typeof this.props.addCustomProjectCustomTestAssetFilesModalUploadPercentage === 'number'
              ? () => {
                if (currentFileUploadCancelToken === null) return
                Logger.debug('editCustomProjectSite...changeTestView', 'canceled file upload')
                currentFileUploadCancelToken.cancel('canceled')
              }
              : null
            }


            filterText={this.props.addCustomProjectCustomTestAssetFilesModalFilterText}
            onFilterTextChanged={(text) => {
              this.props.setAddCustomProjectCustomTestAssetFilesModalFilterText(text)
            }}


            notDetachableAssetFilePreviews={isCompareFilesTest
                                            ? this.props.test.files.filter(
                p => fileCompares.some(k => k.file1NameWithExtension === p.originalName))
                                            : []}
            notDetachableReason={getI18n(this.props.langId,
                                         "This file cannot be removed because it is part of the compare files test"
            )}

            attachedAssetFilePreviews={this.props.test.files}
            allAssetFilePreviews={this.props.allCustomProjectAssetFilePreviews}

            onAttachAssetFilePreview={async (assetFilePreview) => {
//we don't download the asset file so we don't need a progress bar (when adding an markdown file we need to download it after attaching)

              await this.props.attachExitingFileToCustomProjectCustomTestAsync(this.props.customProjectId,
                                                                               this.props.test.id,
                                                                               assetFilePreview.id,
                                                                               // this.props.setAddAssetFilesModalUploadPercentage
              )
            }}

            onDetachAssetFilePreview={async (assetFilePreview) => {
              await this.props.detachFileFromCustomProjectCustomTestAsync(this.props.customProjectId,
                                                                          this.props.test.id, assetFilePreview.id
              )

              //refresh list
              await this.props.getPossibleAllUserFilesAsync()
            }}

            onClose={() => {
              this.props.setIsAddCustomProjectCustomTestAssetFilesModalDisplayed(false, false)
            }}

          />

          <Form.Group widths='equal'>
            <Form.Field required>
              <label>
                {
                  getI18n(this.props.langId, 'Type')
                }
              </label>
              <MaterialSelectWrapper
                value={this.props.test.testTypeId}
                validations={validationRules.testTypeId}
                validationMessageKey={validationMessageKeys.testTypeId}
              >
                <select
                  value={this.props.test.testTypeId}
                  onChange={(e: SelectOnChangeEventData) => {
                    const id = parseId(e.currentTarget.value)
                    if (id === null) {
                      errorDialog(getI18n(this.props.langId, 'Error'),
                                  getI18n(this.props.langId, 'Could not convert string to int'), this.props.langId
                      )
                    } else {
                      this.props.setTestTypeId(id)
                    }

                  }}
                >
                  {
                    this.props.testTypes.map(testType => {
                      return (
                        <option key={testType.id} value={testType.id}>
                          {
                            testType.displayName
                          }
                        </option>
                      )
                    })
                  }
                </select>
              </MaterialSelectWrapper>

            </Form.Field>

            <Form.Field required>
              <label>
                {
                  getI18n(this.props.langId, 'Points')
                }
              </label>
              <MaterialInput type="number"
                             validations={validationRules.weight}
                             validationMessageKey={validationMessageKeys.weight}
                             value={this.props.test.weight}
                             onChange={(e) => {

                               if (e.currentTarget.value === '') {
                                 this.props.setWeight(0)
                                 return
                               }

                               const weight = parseInt(e.currentTarget.value, 10)

                               if (isNaN(weight) || weight < 0) {

                                 errorDialog(getI18n(this.props.langId, 'Error'),
                                             getI18n(this.props.langId, 'Could not convert string to int'),
                                             this.props.langId
                                 )

                                 return
                               }

                               this.props.setWeight(weight)
                             }}
              />
            </Form.Field>

            <Form.Field required>
              <label>
                {
                  getI18n(this.props.langId, 'Name')
                }
              </label>
              <MaterialInput
                validations={validationRules.displayName}
                validationMessageKey={validationMessageKeys.displayName}
                value={this.props.test.displayName}
                onChange={(e) => {
                  this.props.setDisplayNameAction(e.currentTarget.value)
                }}
              />
            </Form.Field>
          </Form.Group>

        </div>

        <Tab menu={{
          secondary: true,
          pointing: true
        }}
             className="fh flexed-h"
             panes={
               [
                 {
                   menuItem: `${getI18n(this.props.langId, 'Test content')}*`,
                   render: () => {
                     return (
                       <div id={ChangeTestViewResizeId} className="flexed-h flex-stretch-0">
                         {
                           //if we got a parse error
                           (isRegexTest && isNotEmptyOrSpaces(
                             testParseError)) &&
                           <div>
                             {
                               getI18n(this.props.langId, 'Parse Error') + ':'
                             }
                             <pre className="error-colored">
                          {
                            testParseError
                          }
                        </pre>
                           </div>
                         }

                         {
                           //if we got a regex error
                           (isRegexTest && isNotEmptyOrSpaces(
                             regexError)) &&
                           <div>
                             {
                               getI18n(this.props.langId, 'Regex Error') + ':'
                             }
                             <pre className="error-colored">
                          {
                            regexError
                          }
                        </pre>
                           </div>
                         }

                         {
                           isCompareFilesTest &&
                           <div>
                             <CompareFilesTestSettingsView files={this.props.test.files}
                                                           testContent={this.props.test.content}
                                                           setTestContent={this.props.setEditorTestContent}
                                                           isReadonly={false}
                             />
                           </div>
                         }

                         {
                           //calc with because SimpleVDivider has 1em margin
                           //don't conditional render the editor because this probably (?) breaks ace editor storage
                         }
                         <div className="fh flexed-h flex-stretch">
                           <div className="flexed-h flex-stretch">
                             <SimpleCodeEditorWrapper
                               id={changeTestViewCustomProjectEditorId}
                               editorTab={tempEditorTab}
                               onBlur={(value: string) => {
                                 this.props.setEditorTestContent(value)
                               }}
                               debounceTimeInMs={descriptionEditorDebounceInMs}
                               // onChange={(editorTab: EditorTab) => {
                               //   this.props.setEditorTestContent(editorTab.content)
                               // }}
                               // onEditorSessionChanged={null}
                             />
                           </div>

                         </div>

                       </div>
                     )
                   }
                 },
                 {
                   menuItem: getI18n(this.props.langId, 'Test content visually'),
                   render: () => {
                     return (
                       <div className="mar-bot">
                         <ConsolePanelView test={null} protocolItems={visualTest} testTypes={this.props.testTypes}
                                           displayFilterConsoleItemsOptions={false}
                         />
                       </div>
                     )
                   }
                 },
                 {
                   menuItem: getI18n(this.props.langId, 'Test settings & files'),
                   //DISABLED until we find a way to limit upload size (at backend)
                   //div-disabled
                   render: () => {
                     return (
                       //div-disabled
                       <div className="ui form divided-checkboxes fh">


                         <TestSettingsView
                           testTypeLimits={TestTypeLimits.IsCustomProjectTest}
                           timeoutInMs={this.props.test.testSettings.timeoutInMs}
                           memoryLimitInKb={this.props.test.testSettings.memoryLimitInKb}
                           maxDiskSpaceInKb={this.props.test.testSettings.maxDiskSpaceInKb}
                           compileTimeoutInMs={this.props.test.testSettings.compileTimeoutInMs}
                           compilerOptions={this.props.test.testSettings.compilerOptions}

                           set_timeoutInMs={this.props.setEditorTest_timeoutInMs}
                           set_memoryLimitInKb={this.props.setEditorTest_memoryLimitInKb}
                           set_maxDiskSpaceInKb={this.props.setEditorTest_maxDiskSpaceInKb}
                           set_compileTimeoutInMs={this.props.setEditorTest_compileTimeoutInMs}
                           set_compilerOptions={this.props.setEditorTest_compilerOptions}
                         />


                         <h4 className="ui dividing header">
                           {
                             getI18n(this.props.langId, 'Files')
                           }
                         </h4>

                         <div className="flexed">
                           <div className={this.props.test.id > 0
                                           ? 'clickable'
                                           : 'div-disabled'} style={{marginBottom: '0.5em'}}
                                onClick={() => {

                                  if (this.props.test.id <= 0) return

                                  this.props.setIsAddCustomProjectCustomTestAssetFilesModalDisplayed(true, false) //before loading because this will cancel isBusy
                                  this.props.getPossibleAllUserFilesAsync()
                                }}
                           >
                             <Icon name="write"/>
                             <span>
                              {
                                getI18n(this.props.langId, 'Manage')
                              }
                            </span>
                           </div>


                           {
                             this.props.test.id <= 0 &&
                             <HelpPopup className="mar-left-half" defaultText={getI18n(this.props.langId,
                                                                                       "Managing test asset files is only allowed for saved tests (not for new tests)"
                             )}/>
                           }

                           <HelpPopup className="mar-left-half warning-colored" defaultText={getI18n(this.props.langId,
                                                                                                     "All changes to assets are applied immediately in the backend (saving is not needed for this)!"
                           )}/>

                         </div>

                         <TestFilesListView files={this.props.test.files}
                                            getDownloadFileHRef={(file) => {
                                              return `${RequestApiPrefix}assets/custom/project/test/${this.props.customProjectId}/${this.props.test.id}/${file.id}`
                                            }}
                         />
                       </div>
                     )
                   }
                 },
               ]

             }/>


      </div>
    )
  }
}

export default connect(mapStateToProps, mapDispatchToProps)(ChangeTestView)
